@@ -38,15 +38,20 @@ from enum import Enum
 import logging
 
 from fim.slivers.attached_components import ComponentSliver, AttachedComponentsInfo
-from fim.slivers.capacities_labels import Capacities, Labels, ReservationInfo, StructuralInfo, CapacityHints, Location
+from fim.slivers.capacities_labels import Capacities, Labels, ReservationInfo, \
+    StructuralInfo, CapacityHints, Location, Flags
 from fim.slivers.delegations import Delegations, DelegationType
 from fim.slivers.interface_info import InterfaceSliver, InterfaceInfo
 from fim.slivers.base_sliver import BaseSliver
-from fim.slivers.network_node import NodeSliver
+from fim.slivers.network_node import NodeSliver, CompositeNodeSliver
 from fim.slivers.network_link import NetworkLinkSliver
 from fim.slivers.path_info import PathInfo, ERO
-from fim.slivers.network_service import NetworkServiceSliver, NetworkServiceInfo
+from fim.slivers.tags import Tags
+from fim.slivers.json_data import MeasurementData, UserData, LayoutData
+from fim.slivers.network_service import NetworkServiceSliver, NetworkServiceInfo, NSLayer, MirrorDirection
 from fim.graph.abc_property_graph_constants import ABCPropertyGraphConstants
+from fim.slivers.gateway import Gateway
+from fim.slivers.maintenance_mode import MaintenanceInfo
 
 
 class GraphFormat(Enum):
@@ -91,7 +96,18 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
         "structural_info": ABCPropertyGraphConstants.PROP_STRUCTURAL_INFO,
         "ero": ABCPropertyGraphConstants.PROP_ERO,
         "path_info": ABCPropertyGraphConstants.PROP_PATH_INFO,
-        "controller_url": ABCPropertyGraphConstants.PROP_CONTROLLER_URL
+        "controller_url": ABCPropertyGraphConstants.PROP_CONTROLLER_URL,
+        "gateway": ABCPropertyGraphConstants.PROP_GATEWAY,
+        "mirror_port": ABCPropertyGraphConstants.PROP_MIRROR_PORT,
+        "mirror_direction": ABCPropertyGraphConstants.PROP_MIRROR_DIRECTION,
+        "peer_labels": ABCPropertyGraphConstants.PROP_PEER_LABELS,
+        "mf_data": ABCPropertyGraphConstants.PROP_MEAS_DATA,
+        "layout_data": ABCPropertyGraphConstants.PROP_LAYOUT_DATA,
+        "user_data": ABCPropertyGraphConstants.PROP_USER_DATA,
+        "tags": ABCPropertyGraphConstants.PROP_TAGS,
+        "flags": ABCPropertyGraphConstants.PROP_FLAGS,
+        "boot_script": ABCPropertyGraphConstants.PROP_BOOT_SCRIPT,
+        "maintenance_info": ABCPropertyGraphConstants.PROP_MAINTENANCE_INFO
     }
 
     @abstractmethod
@@ -466,38 +482,56 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
         """
         prop_dict = dict()
 
-        if sliver.resource_name is not None:
+        # hasattr checks are added because slivers sometimes are unpickled and certain
+        # attributes may not be present (come from an older version of the code)
+        if hasattr(sliver, 'resource_name') and sliver.resource_name is not None:
             prop_dict[ABCPropertyGraph.PROP_NAME] = sliver.resource_name
-        if sliver.resource_type is not None:
+        if hasattr(sliver, 'resource_type') and sliver.resource_type is not None:
             prop_dict[ABCPropertyGraph.PROP_TYPE] = str(sliver.resource_type)
-        if sliver.resource_model is not None:
+        if hasattr(sliver, 'resource_model') and sliver.resource_model is not None:
             prop_dict[ABCPropertyGraph.PROP_MODEL] = sliver.resource_model
-        if sliver.capacities is not None:
+        if hasattr(sliver, 'capacities') and sliver.capacities is not None:
             prop_dict[ABCPropertyGraph.PROP_CAPACITIES] = sliver.capacities.to_json()
-        if sliver.capacity_hints is not None:
+        if hasattr(sliver, 'capacity_hints') and sliver.capacity_hints is not None:
             prop_dict[ABCPropertyGraph.PROP_CAPACITY_HINTS] = sliver.capacity_hints.to_json()
-        if sliver.labels is not None:
+        if hasattr(sliver, 'labels') and sliver.labels is not None:
             prop_dict[ABCPropertyGraph.PROP_LABELS] = sliver.labels.to_json()
-        if sliver.capacity_delegations is not None:
+        if hasattr(sliver, 'capacity_delegations') and sliver.capacity_delegations is not None:
             prop_dict[ABCPropertyGraph.PROP_CAPACITY_DELEGATIONS] = \
                 sliver.capacity_delegations.to_json()
-        if sliver.label_delegations is not None:
+        if hasattr(sliver, 'label_delegations') and sliver.label_delegations is not None:
             prop_dict[ABCPropertyGraph.PROP_LABEL_DELEGATIONS] = \
                 sliver.label_delegations.to_json()
-        if sliver.capacity_allocations is not None:
+        if hasattr(sliver, 'capacity_allocations') and sliver.capacity_allocations is not None:
             prop_dict[ABCPropertyGraph.PROP_CAPACITY_ALLOCATIONS] = sliver.capacity_allocations.to_json()
-        if sliver.label_allocations is not None:
+        if hasattr(sliver, 'label_allocations') and sliver.label_allocations is not None:
             prop_dict[ABCPropertyGraph.PROP_LABEL_ALLOCATIONS] = sliver.label_allocations.to_json()
-        if sliver.reservation_info is not None:
+        if hasattr(sliver, 'reservation_info') and sliver.reservation_info is not None:
             prop_dict[ABCPropertyGraph.PROP_RESERVATION_INFO] = sliver.reservation_info.to_json()
-        if sliver.structural_info is not None:
+        if hasattr(sliver, 'structural_info') and sliver.structural_info is not None:
             prop_dict[ABCPropertyGraph.PROP_STRUCTURAL_INFO] = sliver.structural_info.to_json()
-        if sliver.details is not None:
+        if hasattr(sliver, 'details') and sliver.details is not None:
             prop_dict[ABCPropertyGraph.PROP_DETAILS] = sliver.details
-        if sliver.node_map is not None:
+        if hasattr(sliver, 'node_map') and sliver.node_map is not None:
             prop_dict[ABCPropertyGraph.PROP_NODE_MAP] = json.dumps(sliver.node_map)
-        # boolean is always there. use json dumps for simplicity
-        prop_dict[ABCPropertyGraph.PROP_STITCH_NODE] = json.dumps(sliver.stitch_node)
+        if hasattr(sliver, 'stitch_node'):
+            # boolean is always there. use json dumps for simplicity
+            prop_dict[ABCPropertyGraph.PROP_STITCH_NODE] = json.dumps(sliver.stitch_node)
+        # this is already a JSON dict
+        if hasattr(sliver, 'mf_data') and sliver.mf_data is not None:
+            prop_dict[ABCPropertyGraph.PROP_MEAS_DATA] = sliver.mf_data.json
+        # this is already a JSON dict
+        if hasattr(sliver, 'user_data') and sliver.user_data is not None:
+            prop_dict[ABCPropertyGraph.PROP_USER_DATA] = sliver.user_data.json
+        # this is already a JSON dict
+        if hasattr(sliver, 'layout_data') and sliver.layout_data is not None:
+            prop_dict[ABCPropertyGraph.PROP_LAYOUT_DATA] = sliver.layout_data.json
+        if hasattr(sliver, 'tags') and sliver.tags is not None:
+            prop_dict[ABCPropertyGraph.PROP_TAGS] = sliver.tags.to_json()
+        if hasattr(sliver, 'flags') and sliver.flags is not None:
+            prop_dict[ABCPropertyGraph.PROP_FLAGS] = sliver.flags.to_json()
+        if hasattr(sliver, 'boot_script') and sliver.boot_script is not None:
+            prop_dict[ABCPropertyGraph.PROP_BOOT_SCRIPT] = sliver.boot_script
 
         return prop_dict
 
@@ -510,18 +544,23 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
         """
         prop_dict = ABCPropertyGraph.base_sliver_to_graph_properties_dict(sliver)
 
-        if sliver.image_ref is not None and sliver.image_type is not None:
+        # hasattr checks are added because slivers sometimes are unpickled and certain
+        # attributes may not be present (come from an older version of the code)
+        if hasattr(sliver, 'image_ref') and hasattr(sliver, 'image_type')  and \
+                sliver.image_ref is not None and sliver.image_type is not None:
             prop_dict[ABCPropertyGraph.PROP_IMAGE_REF] = sliver.image_ref + ',' + str(sliver.image_type)
-        if sliver.management_ip is not None:
+        if hasattr(sliver, 'management_ip') and sliver.management_ip is not None:
             prop_dict[ABCPropertyGraph.PROP_MGMT_IP] = str(sliver.management_ip)
-        if sliver.allocation_constraints is not None:
+        if hasattr(sliver, 'allocation_constraints') and sliver.allocation_constraints is not None:
             prop_dict[ABCPropertyGraph.PROP_ALLOCATION_CONSTRAINTS] = sliver.allocation_constraints
-        if sliver.service_endpoint is not None:
+        if hasattr(sliver, 'service_endpoint') and sliver.service_endpoint is not None:
             prop_dict[ABCPropertyGraph.PROP_SERVICE_ENDPOINT] = str(sliver.service_endpoint)
-        if sliver.site is not None:
+        if hasattr(sliver, 'site') and sliver.site is not None:
             prop_dict[ABCPropertyGraph.PROP_SITE] = sliver.site
-        if sliver.location is not None:
+        if hasattr(sliver, 'location') and sliver.location is not None:
             prop_dict[ABCPropertyGraph.PROP_LOCATION] = sliver.location.to_json()
+        if hasattr(sliver, 'maintenance_info') and sliver.maintenance_info is not None:
+            prop_dict[ABCPropertyGraph.PROP_MAINTENANCE_INFO] = sliver.maintenance_info.to_json()
 
         return prop_dict
 
@@ -534,9 +573,11 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
         """
         prop_dict = ABCPropertyGraph.base_sliver_to_graph_properties_dict(sliver)
 
-        if sliver.layer is not None:
+        # hasattr checks are added because slivers sometimes are unpickled and certain
+        # attributes may not be present (come from an older version of the code)
+        if hasattr(sliver, 'layer') and sliver.layer is not None:
             prop_dict[ABCPropertyGraph.PROP_LAYER] = str(sliver.layer)
-        if sliver.technology is not None:
+        if hasattr(sliver, 'technology') and sliver.technology is not None:
             prop_dict[ABCPropertyGraph.PROP_TECHNOLOGY] = str(sliver.technology)
 
         return prop_dict
@@ -560,20 +601,29 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
         :return:
         """
         prop_dict = ABCPropertyGraph.base_sliver_to_graph_properties_dict(sliver)
-        if sliver.layer is not None:
+
+        # hasattr checks are added because slivers sometimes are unpickled and certain
+        # attributes may not be present (come from an older version of the code)
+        if hasattr(sliver, 'layer') and sliver.layer is not None:
             prop_dict[ABCPropertyGraph.PROP_LAYER] = str(sliver.layer)
-        if sliver.technology is not None:
+        if hasattr(sliver, 'technology') and sliver.technology is not None:
             prop_dict[ABCPropertyGraph.PROP_TECHNOLOGY] = str(sliver.technology)
-        if sliver.allocation_constraints is not None:
+        if hasattr(sliver, 'allocation_constraints') and sliver.allocation_constraints is not None:
             prop_dict[ABCPropertyGraph.PROP_ALLOCATION_CONSTRAINTS] = sliver.allocation_constraints
-        if sliver.ero is not None:
+        if hasattr(sliver, 'ero') and sliver.ero is not None:
             prop_dict[ABCPropertyGraph.PROP_ERO] = sliver.ero.to_json()
-        if sliver.path_info is not None:
+        if hasattr(sliver, 'path_info') and sliver.path_info is not None:
             prop_dict[ABCPropertyGraph.PROP_PATH_INFO] = sliver.path_info.to_json()
-        if sliver.controller_url is not None:
+        if hasattr(sliver, 'controller_url') and sliver.controller_url is not None:
             prop_dict[ABCPropertyGraph.PROP_CONTROLLER_URL] = sliver.controller_url
-        if sliver.site is not None:
+        if hasattr(sliver, 'site') and sliver.site is not None:
             prop_dict[ABCPropertyGraph.PROP_SITE] = sliver.site
+        if hasattr(sliver, 'gateway') and sliver.gateway is not None:
+            prop_dict[ABCPropertyGraph.PROP_GATEWAY] = sliver.gateway.to_json()
+        if hasattr(sliver, 'mirror_port') and sliver.mirror_port is not None:
+            prop_dict[ABCPropertyGraph.PROP_MIRROR_PORT] = sliver.mirror_port
+        if hasattr(sliver, 'mirror_direction') and sliver.mirror_direction is not None:
+            prop_dict[ABCPropertyGraph.PROP_MIRROR_DIRECTION] = str(sliver.mirror_direction)
 
         return prop_dict
 
@@ -586,8 +636,67 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
         """
         prop_dict = ABCPropertyGraph.base_sliver_to_graph_properties_dict(sliver)
 
+        if hasattr(sliver, 'peer_labels') and sliver.peer_labels is not None:
+            prop_dict[ABCPropertyGraph.PROP_PEER_LABELS] = sliver.peer_labels.to_json()
+
         return prop_dict
 
+    @staticmethod
+    def sliver_to_dict(sliver) -> Dict[str, Any]:
+        """
+        Convert sliver to a deep dictionary. .
+        :param sliver:
+        :return:
+        """
+        assert (issubclass(type(sliver), BaseSliver))
+
+        # convert to properties dict first
+        if type(sliver) == NodeSliver or type(sliver) == CompositeNodeSliver:
+            d = ABCPropertyGraph.node_sliver_to_graph_properties_dict(sliver)
+            # now add deep sliver stuff
+            # components and network services
+            if sliver.attached_components_info is not None:
+                ac = list()
+                for c in sliver.attached_components_info.list_devices():
+                    ac.append(ABCPropertyGraph.sliver_to_dict(c))
+                d['components'] = ac
+
+            if sliver.network_service_info is not None:
+                nss = list()
+                for ns in sliver.network_service_info.list_network_services():
+                    nss.append(ABCPropertyGraph.sliver_to_dict(ns))
+                d['network_services'] = nss
+
+        elif type(sliver) == NetworkServiceSliver:
+            d = ABCPropertyGraph.network_service_sliver_to_graph_properties_dict(sliver)
+            # now add deep sliver stuff
+            # interfaces
+            if sliver.interface_info is not None:
+                ii = list()
+                for i in sliver.interface_info.list_interfaces():
+                    ii.append(ABCPropertyGraph.sliver_to_dict(i))
+                d['interfaces'] = ii
+
+        elif type(sliver) == ComponentSliver:
+            d = ABCPropertyGraph.component_sliver_to_graph_properties_dict(sliver)
+            # now add deep sliver stuff
+            # network services
+            if sliver.network_service_info is not None:
+                nss = list()
+                for ns in sliver.network_service_info.list_services():
+                    nss.append(ABCPropertyGraph.sliver_to_dict(ns))
+                d['network_services'] = nss
+
+        elif type(sliver) == NetworkLinkSliver:
+            d = ABCPropertyGraph.link_sliver_to_graph_properties_dict(sliver)
+        elif type(sliver) == InterfaceSliver:
+            d = ABCPropertyGraph.interface_sliver_to_graph_properties_dict(sliver)
+        else:
+            raise PropertyGraphQueryException(msg=f'JSON Conversion for type {type(sliver)} is not supported.',
+                                              graph_id=None, node_id=None)
+
+        return d
+
     @staticmethod
     def set_base_sliver_properties_from_graph_properties_dict(sliver: BaseSliver, d: Dict[str, str]):
         """
@@ -625,7 +734,19 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
                               node_map=json.loads(d[ABCPropertyGraph.PROP_NODE_MAP]) if
                               d.get(ABCPropertyGraph.PROP_NODE_MAP, None) is not None else None,
                               stitch_node=json.loads(d[ABCPropertyGraph.PROP_STITCH_NODE]) if
-                              d.get(ABCPropertyGraph.PROP_STITCH_NODE, None) is not None else False
+                              d.get(ABCPropertyGraph.PROP_STITCH_NODE, None) is not None else False,
+                              tags=Tags.from_json(d.get(ABCPropertyGraph.PROP_TAGS, None)),
+                              flags=Flags.from_json(d.get(ABCPropertyGraph.PROP_FLAGS, None)),
+                              mf_data=MeasurementData(d[ABCPropertyGraph.PROP_MEAS_DATA])
+                                                      if d.get(ABCPropertyGraph.PROP_MEAS_DATA, None)
+                                                         is not None else None,
+                              user_data=UserData(d[ABCPropertyGraph.PROP_USER_DATA])
+                                                  if d.get(ABCPropertyGraph.PROP_USER_DATA, None)
+                                                     is not None else None,
+                              layout_data=LayoutData(d[ABCPropertyGraph.PROP_LAYOUT_DATA])
+                                                  if d.get(ABCPropertyGraph.PROP_LAYOUT_DATA, None)
+                                                     is not None else None,
+                              boot_script=d.get(ABCPropertyGraph.PROP_BOOT_SCRIPT, None)
                               )
 
     @staticmethod
@@ -643,7 +764,9 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
                          allocation_constraints=d.get(ABCPropertyGraph.PROP_ALLOCATION_CONSTRAINTS, None),
                          service_endpoint=d.get(ABCPropertyGraph.PROP_SERVICE_ENDPOINT, None),
                          site=d.get(ABCPropertyGraphConstants.PROP_SITE, None),
-                         location=Location.from_json(d.get(ABCPropertyGraph.PROP_LOCATION, None))
+                         location=Location.from_json(d.get(ABCPropertyGraph.PROP_LOCATION, None)),
+                         maintenance_info=
+                         MaintenanceInfo.from_json(d.get(ABCPropertyGraph.PROP_MAINTENANCE_INFO, None))
                          )
         return n
 
@@ -651,7 +774,7 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
     def link_sliver_from_graph_properties_dict(d: Dict[str, str]) -> NetworkLinkSliver:
         n = NetworkLinkSliver()
         ABCPropertyGraph.set_base_sliver_properties_from_graph_properties_dict(n, d)
-        n.set_properties(layer=d.get(ABCPropertyGraph.PROP_LAYER),
+        n.set_properties(layer=NSLayer.from_string(d.get(ABCPropertyGraph.PROP_LAYER)),
                          technology=d.get(ABCPropertyGraph.PROP_TECHNOLOGY),
                          )
         return n
@@ -676,13 +799,16 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
         """
         ns = NetworkServiceSliver()
         ABCPropertyGraph.set_base_sliver_properties_from_graph_properties_dict(ns, d)
-        ns.set_properties(layer=d.get(ABCPropertyGraph.PROP_LAYER),
+        ns.set_properties(layer=NSLayer.from_string(d.get(ABCPropertyGraph.PROP_LAYER)),
                           technology=d.get(ABCPropertyGraph.PROP_TECHNOLOGY),
                           allocation_constraints=d.get(ABCPropertyGraph.PROP_ALLOCATION_CONSTRAINTS, None),
                           ero=ERO.from_json(d.get(ABCPropertyGraph.PROP_ERO, None)),
                           path_info=PathInfo.from_json(d.get(ABCPropertyGraph.PROP_PATH_INFO, None)),
                           controller_url=d.get(ABCPropertyGraph.PROP_CONTROLLER_URL, None),
-                          site=d.get(ABCPropertyGraphConstants.PROP_SITE, None)
+                          site=d.get(ABCPropertyGraphConstants.PROP_SITE, None),
+                          gateway=Gateway.from_json(d.get(ABCPropertyGraphConstants.PROP_GATEWAY, None)),
+                          mirror_port=d.get(ABCPropertyGraphConstants.PROP_MIRROR_PORT, None),
+                          mirror_direction=MirrorDirection.from_string(d.get(ABCPropertyGraphConstants.PROP_MIRROR_DIRECTION, None))
                           )
         return ns
 
@@ -695,6 +821,7 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
         """
         isl = InterfaceSliver()
         ABCPropertyGraph.set_base_sliver_properties_from_graph_properties_dict(isl, d)
+        isl.set_properties(peer_labels=Labels.from_json(d.get(ABCPropertyGraph.PROP_PEER_LABELS, None)))
         return isl
 
     def build_deep_node_sliver(self, *, node_id: str) -> NodeSliver:
@@ -731,6 +858,35 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
 
         return ns
 
+    @staticmethod
+    def build_deep_node_sliver_from_dict(*, props: Dict[str, Any]) -> NodeSliver:
+        """
+        Build a deep NetworkNode or other similar (e.g.
+        network-attached storage) sliver from a deep dictionary
+        :param props:
+        :return:
+        """
+        # create top-level sliver
+        ns = ABCPropertyGraph.node_sliver_from_graph_properties_dict(props)
+        # find and build deep slivers of network services (if any) and components (if any)
+        comps = props.get('components', None)
+        if comps is not None and len(comps) > 0:
+            aci = AttachedComponentsInfo()
+            for c in comps:
+                cs = ABCPropertyGraph.build_deep_component_sliver_from_dict(props=c)
+                aci.add_device(cs)
+            ns.attached_components_info = aci
+
+        nss = props.get('network_services', None)
+        if nss is not None and len(nss) > 0:
+            nsi = NetworkServiceInfo()
+            for s in nss:
+                nssl = ABCPropertyGraph.build_deep_ns_sliver_from_dict(props=s)
+                nsi.add_network_service(nssl)
+            ns.network_service_info = nsi
+
+        return ns
+
     def build_deep_ns_sliver(self, *, node_id: str) -> NetworkServiceSliver:
         """
         Build a deep sliver for a NetworkService
@@ -755,7 +911,31 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
             nss.interface_info = ifi
         return nss
 
+    @staticmethod
+    def build_deep_ns_sliver_from_dict(*, props: Dict[str, Any]) -> NetworkServiceSliver:
+        """
+        Build a deep sliver for a NetworkService from a deep dictionary
+        :param props:
+        :return:
+        """
+        # create top-level sliver
+        nss = ABCPropertyGraph.network_service_sliver_from_graph_properties_dict(props)
+        # find interfaces and attach
+        ifs = props.get('interfaces', None)
+        if ifs is not None and len(ifs) > 0:
+            ifi = InterfaceInfo()
+            for i in ifs:
+                ifsl = ABCPropertyGraph.interface_sliver_from_graph_properties_dict(i)
+                ifi.add_interface(ifsl)
+            nss.interface_info = ifi
+        return nss
+
     def build_deep_component_sliver(self, *, node_id: str) -> ComponentSliver:
+        """
+        Build a deep component slliver from graph
+        :param node_id:
+        :return:
+        """
         clazzes, props = self.get_node_properties(node_id=node_id)
         if ABCPropertyGraph.CLASS_Component not in clazzes:
             raise PropertyGraphQueryException(node_id=node_id, graph_id=self.graph_id,
@@ -773,6 +953,75 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
             cs.network_service_info = nsi
         return cs
 
+    @staticmethod
+    def build_deep_component_sliver_from_dict(*, props: Dict[str, Any]) -> ComponentSliver:
+        """
+        Build a deep component sliver from deep dictionary
+        :param props:
+        :return:
+        """
+        # create top-level sliver
+        cs = ABCPropertyGraph.component_sliver_from_graph_properties_dict(props)
+        # find any network services, build and attach
+        nss = props.get('network_services', None)
+        if nss is not None and len(nss) > 0:
+            nsi = NetworkServiceInfo()
+            for s in nss:
+                nssl = ABCPropertyGraph.build_deep_ns_sliver_from_dict(props=s)
+                nsi.add_network_service(nssl)
+            cs.network_service_info = nsi
+        return cs
+
+    def build_deep_interface_sliver(self, *, node_id: str) -> InterfaceSliver:
+        """
+        Build a deep interface sliver from graph
+        :param node_id:
+        :return:
+        """
+        clazzes, props = self.get_node_properties(node_id=node_id)
+        if ABCPropertyGraph.CLASS_ConnectionPoint not in clazzes:
+            raise PropertyGraphQueryException(node_id=node_id, graph_id=self.graph_id,
+                                              msg="Node is not of class Interface")
+        # create top-level sliver
+        isl = ABCPropertyGraph.interface_sliver_from_graph_properties_dict(props)
+        return isl
+
+    @staticmethod
+    def build_deep_interface_sliver_from_dict(*, props: Dict[str, Any]) -> InterfaceSliver:
+        """
+        Build a deep interface sliver from graph
+        :param props:
+        :return:
+        """
+        # create top-level sliver
+        isl = ABCPropertyGraph.interface_sliver_from_graph_properties_dict(props)
+        return isl
+
+    def build_deep_link_sliver(self, *, node_id: str) -> NetworkLinkSliver:
+        """
+        Build a deep link sliver from graph
+        :param node_id:
+        :return:
+        """
+        clazzes, props = self.get_node_properties(node_id=node_id)
+        if ABCPropertyGraph.CLASS_Link not in clazzes:
+            raise PropertyGraphQueryException(node_id=node_id, graph_id=self.graph_id,
+                                              msg="Node is not of class Link")
+        # create top-level sliver
+        lsl = ABCPropertyGraph.link_sliver_from_graph_properties_dict(props)
+        return lsl
+
+    @staticmethod
+    def build_deep_link_sliver_from_dict(*, props: Dict[str, Any]) -> NetworkLinkSliver:
+        """
+        Build a deep link sliver from graph
+        :param props:
+        :return:
+        """
+        # create top-level sliver
+        lsl = ABCPropertyGraph.link_sliver_from_graph_properties_dict(props)
+        return lsl
+
     def remove_network_node_with_components_nss_cps_and_links(self, node_id: str):
         """
         Remove a network node, all of components and their interfaces, parent interfaces
@@ -897,7 +1146,8 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
 
         if not self.check_node_unique(label=ABCPropertyGraph.CLASS_NetworkNode,
                                       name=sliver.resource_name):
-            raise RuntimeError(f'Node name {sliver.resource_name} must be unique.')
+            raise PropertyGraphQueryException(msg=f'Node name {sliver.resource_name} must be unique.',
+                                              graph_id=self.graph_id, node_id=None)
 
         props = self.node_sliver_to_graph_properties_dict(sliver)
         self.add_node(node_id=sliver.node_id, label=ABCPropertyGraph.CLASS_NetworkNode, props=props)
@@ -956,7 +1206,8 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
         if parent_node_id is None and not self.check_node_unique(label=ABCPropertyGraph.CLASS_NetworkService,
                                                                  name=network_service.resource_name):
             # slice-wide network services must have unique names
-            raise RuntimeError(f'Network service name {network_service.resource_name} must be unique.')
+            raise PropertyGraphQueryException(msg=f'Network service name {network_service.resource_name} must be unique.',
+                                              graph_id=self.graph_id, node_id=parent_node_id)
 
         props = self.network_service_sliver_to_graph_properties_dict(network_service)
         self.add_node(node_id=network_service.node_id, label=ABCPropertyGraph.CLASS_NetworkService, props=props)
@@ -970,17 +1221,21 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
 
     def add_interface_sliver(self, *, parent_node_id: str, interface: InterfaceSliver):
         """
-        Add interface to a network service, linking back to parent.
-        :param parent_node_id: network service id
+        Add interface to a network service, linking back to parent (node or network service),
+        if provided.
+        For most interface slivers there should be a parent with the exception
+        of FacilityPorts/StitchPorts.
+
+        :param parent_node_id: network service id or other parent if available
         :param interface: interface sliver description
         :return:
         """
         assert interface.node_id is not None
-        assert parent_node_id is not None
 
         props = self.interface_sliver_to_graph_properties_dict(interface)
         self.add_node(node_id=interface.node_id, label=ABCPropertyGraph.CLASS_ConnectionPoint, props=props)
-        self.add_link(node_a=parent_node_id, rel=ABCPropertyGraph.REL_CONNECTS, node_b=interface.node_id)
+        if parent_node_id is not None:
+            self.add_link(node_a=parent_node_id, rel=ABCPropertyGraph.REL_CONNECTS, node_b=interface.node_id)
 
     def get_all_ns_or_link_connection_points(self, link_id: str) -> List[str]:
         """
@@ -1006,9 +1261,10 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
         assert parent_node_id is not None
         labels, parent_props = self.get_node_properties(node_id=parent_node_id)
         if ABCPropertyGraph.CLASS_NetworkNode not in labels and \
-                ABCPropertyGraph.CLASS_Component not in labels:
+                ABCPropertyGraph.CLASS_Component not in labels and \
+                ABCPropertyGraph.CLASS_CompositeNode not in labels:
             raise PropertyGraphQueryException(graph_id=self.graph_id, node_id=parent_node_id,
-                                              msg="Parent node type is not NetworkNode or Component")
+                                              msg="Parent node type is not NetworkNode, CompositeNode or Component")
         nss_ifs = self.get_first_and_second_neighbor(node_id=parent_node_id, rel1=ABCPropertyGraph.REL_HAS,
                                                      node1_label=ABCPropertyGraph.CLASS_NetworkService,
                                                      rel2=ABCPropertyGraph.REL_CONNECTS,
@@ -1029,9 +1285,9 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
         :param parent: parent class
         :return:
         """
-        assert node_id is not None
-        assert rel is not None
-        assert parent is not None
+        #assert node_id is not None
+        #assert rel is not None
+        #assert parent is not None
         parent_ids = self.get_first_neighbor(node_id=node_id, rel=rel,
                                              node_label=parent)
         if len(parent_ids) != 1:
@@ -1047,6 +1303,42 @@ class ABCPropertyGraph(ABCPropertyGraphConstants):
         :return:
         """
 
+    @abstractmethod
+    def get_graph_diff(self, other_graph, label: str):
+        """
+        Return two lists nodes (with all properties) that are in this graph but NOT in the other graph
+        and node that are in the other graph, but not in this graph, using label
+        as a filter
+        [0] - elements present in self, absent in other (i.e. removed elements)
+        [1] - elements present in other, absent in self (i.e. added elements
+        """
+
+    @abstractmethod
+    def get_graph_property_diff(self, other_graph, label: str):
+        """
+        Return two lists of nodes (with all properties, from both graphs) that are present in both
+        this and other graph where the several specific properties are different
+        """
+
+    def find_peer_connection_points(self, *, node_id: str) -> List[str] or None:
+        """
+        Find the ids of the peer connection points to this one (connected over a Link)
+        if they exist
+        :param node_id: id of the interface/connection point
+        :return:
+        """
+        assert node_id is not None
+
+        # find id of connection point over a Link
+        candidates = self.get_first_and_second_neighbor(node_id=node_id, rel1=ABCPropertyGraph.REL_CONNECTS,
+                                                        node1_label=ABCPropertyGraph.CLASS_Link,
+                                                        rel2=ABCPropertyGraph.REL_CONNECTS,
+                                                        node2_label=ABCPropertyGraph.CLASS_ConnectionPoint)
+        if len(candidates) == 0:
+            return None
+
+        return [x[1] for x in candidates]
+
 
 class ABCGraphImporter(ABC):
     @abstractmethod
@@ -1165,7 +1457,23 @@ class ABCGraphImporter(ABC):
         :return:
         """
         assert graph_file is not None
-        g = nx.read_graphml(graph_file)
+        # try to read it using known methods
+        g = None
+        try:
+            g = nx.read_graphml(graph_file)
+        except:
+            pass
+
+        try:
+            with open(graph_file, 'r', encoding='utf-8') as f:
+                data = f.read()
+            data = json.loads(data)
+            g = nx.readwrite.node_link_graph(data=data)
+        except:
+            pass
+
+        if not g:
+            raise PropertyGraphImportException(graph_id=None, msg=f'Unable to read graph {graph_file}')
 
         # check graph_ids on nodes
         graph_ids = set()
