@@ -257,6 +257,7 @@ class Neo4jPropertyGraph(ABCPropertyGraph):
 
         query = f"MATCH (s:GraphNode {{GraphID: $graphId, NodeID: $nodeId}}) " \
             f"SET s+= {{ {all_props} }} RETURN properties(s)"
+
         with self.driver.session() as session:
             val = session.run(query, graphId=self.graph_id, nodeId=node_id)
             if val is None or len(val.value()) == 0:
@@ -602,6 +603,77 @@ class Neo4jPropertyGraph(ABCPropertyGraph):
                 return True
             return False
 
+    def get_graph_diff(self, other_graph, label: str):
+        assert other_graph is not None
+        assert label is not None
+
+        query = f"MATCH(n:{label} {{GraphID: $graphIdA}}) WITH n MATCH(n1:{label} {{GraphID: $graphIdB}}) WITH " \
+                f"collect(DISTINCT n) as A, collect(DISTINCT n1) as B, " \
+                f"collect(DISTINCT n.NodeID) as AN, collect(DISTINCT n1.NodeID) as BN " \
+                f"RETURN [x in A WHERE NOT x.NodeID in BN] as AnotB, [x in B WHERE NOT x.NodeID in AN] as BnotA"
+        with self.driver.session() as session:
+            val = session.run(query, graphIdA=self.graph_id, graphIdB=other_graph.graph_id).single()
+            if val is None:
+                raise PropertyGraphQueryException(graph_id=self.graph_id,
+                                                  node_id=None, msg=f"Unable to diff with graph {other_graph.graph_id}")
+            # A not B means element was removed
+            # B not A means element was added
+            return val.data()['AnotB'], val.data()['BnotA']
+
+    def get_graph_property_diff(self, other_graph, label: str):
+        assert other_graph is not None
+        assert label is not None
+        assert property is not None
+
+        # alternative query
+        # FIXME: figure out which one is faster
+        #query = f"MATCH(n:GraphNode:{label} {{GraphID: $graphIdA}}) WITH n MATCH(n1:GraphNode:{label} {{GraphID: $graphIdB}}) " \
+        #        f"WHERE n.NodeID = n1.NodeID AND " \
+        #        f"(" \
+        #        f"((exists(n.Labels) AND exists(n1.Labels) AND n.Labels <> n1.Labels) OR " \
+        #       f"(exists(n.Labels) AND NOT exists(n1.Labels)) OR " \
+        #        f"(NOT exists(n.Labels) AND exists(n1.Labels))) " \
+        #        f"OR " \
+        #        f"((exists(n.Capacities) AND exists(n1.Capacities) AND n.Capacities <> n1.Capacities) OR " \
+        #        f"(exists(n.Capacities) AND NOT exists(n1.Capacities)) OR " \
+        #        f"(NOT exists(n.Capacities) AND exists(n1.Capacities))) " \
+        #        f"OR " \
+        #        f"((exists(n.UserData) AND exists(n1.UserData) AND n.UserData <> n1.UserData) OR " \
+        #        f"(exists(n.UserData) AND NOT exists(n1.UserData)) OR " \
+        #        f"(NOT exists(n.UserData) AND exists(n1.UserData))) " \
+        #        f") " \
+        #        f"RETURN collect(n) as nodes"
+
+        query = f"MATCH(n:GraphNode:{label} {{GraphID: $graphIdA}}) WITH n MATCH(n1:GraphNode:{label} {{GraphID: $graphIdB}}) WHERE " \
+                f"n.NodeID = n1.NodeID AND " \
+                f"((exists(n.Labels) AND exists(n1.Labels) AND n.Labels <> n1.Labels) OR " \
+                f"(exists(n.Labels) AND NOT exists(n1.Labels)) OR " \
+                f"(NOT exists(n.Labels) AND exists(n1.Labels))) " \
+                f"RETURN collect(n) as nodes, collect(n1) as nodes1 " \
+                f"UNION " \
+                f"MATCH(n:GraphNode:{label} {{GraphID: $graphIdA}}) WITH n MATCH(n1:GraphNode:{label} {{GraphID: $graphIdB}}) WHERE " \
+                f"n.NodeID = n1.NodeID AND " \
+                f"((exists(n.Capacities) AND exists(n1.Capacities) AND n.Capacities <> n1.Capacities) OR " \
+                f"(exists(n.Capacities) AND NOT exists(n1.Capacities)) OR " \
+                f"(NOT exists(n.Capacities) AND exists(n1.Capacities))) " \
+                f"RETURN collect(n) as nodes, collect(n1) as nodes1 " \
+                f"UNION " \
+                f"MATCH(n:GraphNode:{label} {{GraphID: $graphIdA}}) WITH n MATCH(n1:GraphNode:{label} {{GraphID: $graphIdB}}) WHERE " \
+                f"n.NodeID = n1.NodeID AND " \
+                f"((exists(n.UserData) AND exists(n1.UserData) AND n.UserData <> n1.UserData) OR " \
+                f"(exists(n.UserData) AND NOT exists(n1.UserData)) OR " \
+                f"(NOT exists(n.UserData) AND exists(n1.UserData))) " \
+                f"RETURN collect(n) as nodes, collect(n1) as nodes1 "
+
+        with self.driver.session() as session:
+            val = session.run(query, graphIdA=self.graph_id, graphIdB=other_graph.graph_id)
+            vd = val.data()
+            if vd is None:
+                raise PropertyGraphQueryException(graph_id=self.graph_id,
+                                                  node_id=None, msg=f"Unable to property diff with graph "
+                                                                    f"{other_graph.graph_id}")
+            return vd[0]['nodes'], vd[0]['nodes1']
+
 
 class Neo4jGraphImporter(ABCGraphImporter):
 
@@ -625,6 +697,8 @@ class Neo4jGraphImporter(ABCGraphImporter):
         self.import_dir = import_dir
         try:
             self.driver = GraphDatabase.driver(self.url, auth=(user, pswd))
+            # per https://neo4j.com/developer-blog/neo4j-driver-best-practices/
+            self.driver.verify_connectivity()
         except Exception as e:
             msg = f"Unable to connect to Neo4j: {str(e)}"
             raise PropertyGraphImportException(graph_id=None, msg=msg)
@@ -634,6 +708,10 @@ class Neo4jGraphImporter(ABCGraphImporter):
             self.log = logger
         self._add_indexes()
 
+    def __del__(self):
+        if self.driver:
+            self.driver.close()
+
     def _add_indexes(self):
         """
         Add required indexes in idempotent manner
@@ -647,13 +725,21 @@ class Neo4jGraphImporter(ABCGraphImporter):
         index_dict = json.load(f)
         f.close()
         self.log.info('Adding Neo4j indexes')
-        for index_name, index_cmd in index_dict.items():
-            with self.driver.session() as session:
+        excount = 0
+        incount = 0
+        with self.driver.session() as session:
+            for index_name, index_cmd in index_dict.items():
+                incount += 1
                 try:
                     session.run(index_cmd)
                 except:
+                    excount += 1
                     # ignore exceptions
                     pass
+        if excount:
+            self.log.warning(f'There were {excount} exceptions when adding {incount} indexes, '
+                             f'most likely because the indexes are '
+                             f'already installed. If in doubt you can use ":schema" command to see them.')
 
     def _prep_graph(self, graph: str, graph_id: str = None) -> Tuple[str, str, str]:
         """
@@ -706,6 +792,8 @@ class Neo4jGraphImporter(ABCGraphImporter):
 
         assert graphml_file is not None
         assert graph_id is not None
+        # delete this graph if it exists
+        self.delete_graph(graph_id=graph_id)
         try:
             with self.driver.session() as session:
                 self.log.debug(f"Loading graph {graph_id} into Neo4j")
@@ -722,6 +810,11 @@ class Neo4jGraphImporter(ABCGraphImporter):
                 query_string = "MATCH (n {GraphID: $graphId }) " \
                                "CALL apoc.create.addLabels([ id(n) ], [ n.Class ]) YIELD node RETURN node"
                 session.run(query_string, graphId=graph_id)
+                # push class labels into class property on relationships
+                self.log.debug(f"Pushing label into Class property on all relationships")
+                query_string = "MATCH (n {GraphID: $graphId }) - [r] - (n1 {GraphID: $graphId }) " \
+                               "CALL apoc.create.setRelProperty(r, 'Class', TYPE(r)) yield rel return rel"
+                session.run(query_string, graphId=graph_id)
         except Exception as e:
             msg = f"Neo4j APOC import error {str(e)}"
             raise PropertyGraphImportException(graph_id=graph_id, msg=msg)
@@ -751,10 +844,10 @@ class Neo4jGraphImporter(ABCGraphImporter):
                 # something in APOC prevents loading sometimes on some platforms
                 self.log.debug(f"Trying to load the file {mapped_file_name}")
                 self._import_graph(mapped_file_name, assigned_id)
-                retry = -1
+                retry = -5
             except PropertyGraphImportException:
                 self.log.warning(f"Transient error, unable to load, deleting and reimporting graph {assigned_id}")
-                retry = retry - 1
+                retry -= 1
                 self.delete_graph(graph_id=assigned_id)
                 # sleep and try again
                 time.sleep(1.0)
@@ -789,30 +882,28 @@ class Neo4jGraphImporter(ABCGraphImporter):
         # get graph id
         graph_id = self.get_graph_id(graph_file=host_file_name)
 
-        # load file
-        try:
-            with self.driver.session() as session:
-                session.run(
-                    'call apoc.import.graphml( $fileName, {batchSize: 10000, '
-                    'readLabels: true, storeNodeIds: true } ) ',
-                    fileName=mapped_file_name).single()
-                # force one common label on all imported nodes
-                self.log.debug(f"Adding GraphNode label to graph {graph_id}")
-                query_string = "MATCH (n {GraphID: $graphId }) SET n:GraphNode"
-                session.run(query_string, graphId=graph_id)
-                # convert class property into a label as well
-                self.log.debug(f"Converting Class property into Neo4j label for all nodes")
-                query_string = "MATCH (n {GraphID: $graphId }) " \
-                               "CALL apoc.create.addLabels([ id(n) ], [ n.Class ]) YIELD node RETURN node"
-                session.run(query_string, graphId=graph_id)
-        except Exception as e:
-            msg = f"Neo4j APOC import error {str(e)}"
-            raise PropertyGraphImportException(graph_id=None, msg=msg)
+        retry = APOC_RETRY_COUNT
+        while retry > 0:
+            try:
+                # something in APOC prevents loading sometimes on some platforms
+                self.log.debug(f"Trying to load the file {mapped_file_name}")
+                self._import_graph(mapped_file_name, graph_id)
+                retry = -5
+            except PropertyGraphImportException:
+                self.log.warning(f"Transient error, unable to load, deleting and reimporting graph {graph_id}")
+                retry -= 1
+                self.delete_graph(graph_id=graph_id)
+                # sleep and try again
+                time.sleep(1.0)
 
         # unlink temp file
         self.log.debug(f"Unlinking temporary file {host_file_name}")
         os.unlink(host_file_name)
 
+        if retry == 0:
+            raise PropertyGraphImportException(graph_id=graph_id,
+                                               msg='Unable to load graph after multiple attempts')
+
         return Neo4jPropertyGraph(graph_id=graph_id, importer=self, logger=self.log)
 
     def import_graph_from_file_direct(self, *, graph_file: str) -> ABCPropertyGraph:
@@ -837,24 +928,7 @@ class Neo4jGraphImporter(ABCGraphImporter):
         graph_id = self.get_graph_id(graph_file=host_file_name)
 
         # load file
-        try:
-            with self.driver.session() as session:
-                session.run(
-                    'call apoc.import.graphml( $fileName, {batchSize: 10000, '
-                    'readLabels: true, storeNodeIds: true } ) ',
-                    fileName=mapped_file_name).single()
-                # force one common label on all imported nodes
-                self.log.debug(f"Adding GraphNode label to graph {graph_id}")
-                query_string = "MATCH (n {GraphID: $graphId }) SET n:GraphNode"
-                session.run(query_string, graphId=graph_id)
-                # convert class property into a label as well
-                self.log.debug(f"Converting Class property into Neo4j label for all nodes")
-                query_string = "MATCH (n {GraphID: $graphId }) " \
-                               "CALL apoc.create.addLabels([ id(n) ], [ n.Class ]) YIELD node RETURN node"
-                session.run(query_string, graphId=graph_id)
-        except Exception as e:
-            msg = f"Neo4j APOC import error {str(e)}"
-            raise PropertyGraphImportException(graph_id=None, msg=msg)
+        self._import_graph(mapped_file_name, graph_id)
 
         # unlink copied file
         self.log.debug(f"Unlinking copied file {host_file_name}")
