@@ -34,17 +34,14 @@ import uuid
 
 from ..abc_property_graph import ABCPropertyGraph, PropertyGraphQueryException
 from fim.slivers.delegations import DelegationType, Pools, Delegations
-from ..typed_tuples import Label, Capacity
 
 
 class ABCARMPropertyGraph(ABCPropertyGraph):
     """
     Interface for an ARM Mixin on top of a property graph
     """
-    DELEGATION_TYPE_TO_CLASS = {DelegationType.LABEL: (ABCPropertyGraph.PROP_LABEL_DELEGATIONS,
-                                                       Label),
-                                DelegationType.CAPACITY: (ABCPropertyGraph.PROP_CAPACITY_DELEGATIONS,
-                                                          Capacity)}
+    DELEGATION_TYPE_TO_PROP = {DelegationType.LABEL: ABCPropertyGraph.PROP_LABEL_DELEGATIONS,
+                               DelegationType.CAPACITY: ABCPropertyGraph.PROP_CAPACITY_DELEGATIONS}
 
     DelegationInfo = recordclass('DelegationInfo', ['graph_id',
                                                     'graph',
@@ -106,10 +103,11 @@ class ABCARMPropertyGraph(ABCPropertyGraph):
             graph.update_node_property(node_id=node_id, prop_name=prop_name,
                                        prop_val=prop_val.to_json())
 
-    def generate_adms(self) -> Dict:
+    def generate_adms(self, delegation_guids: Dict or None = None) -> Dict:
         """
         Generate delegation models from the current ARM graph and return
         the dictionary delegation_id: Neo4jPropertyGraph for generated ADMs
+        :param delegation_guids: - optional dictionary mapping delegation names to guids
         :return:
         """
         self.node_ids = self.list_all_node_ids()
@@ -126,12 +124,18 @@ class ABCARMPropertyGraph(ABCPropertyGraph):
 
         # generate unique graph ids and empty node sets for ADM graphs
         # add stitch nodes as keep nodes to all ADMs
-        delegations_info = {del_id: ABCARMPropertyGraph.DelegationInfo(graph_id=str(uuid.uuid4()),
-                                                        graph=None,
-                                                        keep_nodes=keep_nodes_sets[del_id].union(stitch_nodes),
-                                                        remove_nodes=set())
+        # NOTE: we overwrite delegation GUID if one is available for a specific delegation name (delegation_id)
+        # and otherwise generate a new random GUID
+        delegations_info = {del_id: ABCARMPropertyGraph.DelegationInfo(graph_id=delegation_guids.get(del_id,
+                                                                                                     str(uuid.uuid4()))
+                                                                        if delegation_guids else str(uuid.uuid4()),
+                            graph=None,
+                            keep_nodes=keep_nodes_sets[del_id].union(stitch_nodes),
+                            remove_nodes=set())
                             for del_id in unique_delegation_ids}
 
+        #print(f"Delegations info for graph {self.graph_id}: {delegations_info}")
+
         # At this point we need to write back delegations for specific delegation ID
         # into separate ADM graphs, then determine additional keep nodes and remove
         # unneeded nodes from each.
@@ -142,7 +146,7 @@ class ABCARMPropertyGraph(ABCPropertyGraph):
                 if delegations_by_node.get(node, None) is None:
                     continue
                 for atype in [DelegationType.LABEL, DelegationType.CAPACITY]:
-                    prop_field_name, _ = self.DELEGATION_TYPE_TO_CLASS[atype]
+                    prop_field_name = self.DELEGATION_TYPE_TO_PROP[atype]
                     ds = None
                     if delegations_by_node[node].get(atype, None) is not None:
                         ds = delegations_by_node[node][atype].return_delegations_for_id(del_id)
@@ -152,8 +156,8 @@ class ABCARMPropertyGraph(ABCPropertyGraph):
 
             # in addition to definite keep nodes, we should also keep nodes that
             # a) lie on shortest paths between two connection points we are keeping (to include the links)
-            # b) switch fabrics that connect to connection points
-            # c) network nodes that have switch fabrics
+            # b) network services that connect to connection points
+            # c) network nodes that have network services
             # Determine nodes to delete by doing a difference between all nodes and keep nodes
             # Remove nodes
             # Annotate keep nodes with proper delegation information
@@ -215,7 +219,7 @@ class ABCARMPropertyGraph(ABCPropertyGraph):
         :return:
         """
         assert node_id is not None
-        prop_field_name, _ = self.DELEGATION_TYPE_TO_CLASS[delegation_type]
+        prop_field_name = self.DELEGATION_TYPE_TO_PROP[delegation_type]
         _, props = self.get_node_properties(node_id=node_id)
         if props.get(prop_field_name, None) is not None and \
                 props[prop_field_name] != ABCARMPropertyGraph.NEO4j_NONE:
