@@ -121,7 +121,7 @@ class JSONField(ABC):
 class Capacities(JSONField):
     """
     Implements basic capacity field handling - encoding and decoding
-    from JSON dictionaries of properties
+    from JSON dictionaries of properties. Only ints are allowed.
     """
     UNITS = {'cpu': '', 'unit': '',
              'core': '', 'ram': 'G',
@@ -151,6 +151,7 @@ class Capacities(JSONField):
         for k, v in kwargs.items():
             if v is not None:
                 assert v >= 0
+                assert isinstance(v, int)
             try:
                 # will toss an exception if field is not defined
                 self.__getattribute__(k)
@@ -192,6 +193,8 @@ class Capacities(JSONField):
         return True
 
     def __eq__(self, other):
+        if not other:
+            return False
         assert isinstance(other, Capacities)
         for f, v in self.__dict__.items():
             if v != other.__dict__[f]:
@@ -224,7 +227,6 @@ class Capacities(JSONField):
                 return False
         return True
 
-
     def __str__(self):
         d = self.__dict__.copy()
         for k in self.__dict__:
@@ -331,14 +333,14 @@ class Labels(JSONField):
         'ipv4': (r'(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)',
                  "192.168.1.1"),
         'ipv4_range': (r'(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)-'
-                      r'(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)',
+                       r'(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)',
                        "192.168.1.1-192.168.1.10"),
         'ipv4_subnet': (r'(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)/[\d]{1,2}',
                         "192.168.1.0/24"),
         'ipv6': (r'(?:[a-fA-F0-9]{0,4}:){0,7}[a-fA-F0-9]{0,4}',
                  "2001:0db8:85a3:0000:0000:8a2e:0370:7334"),
         'ipv6_range': (r'(?:[a-fA-F0-9]{0,4}:){0,7}[a-fA-F0-9]{0,4}-(?:[a-fA-F0-9]{0,4}:){0,7}[a-fA-F0-9]{0,4}',
-                      "2001:0db8:85a3:0000:0000:8a2e:0370:7334-2001:0db8:85a3:0000:0000:8a2e:0370:8334"),
+                       "2001:0db8:85a3:0000:0000:8a2e:0370:7334-2001:0db8:85a3:0000:0000:8a2e:0370:8334"),
         # we allow fewer than 128 (and not necessarily 64) bits to be specified, unlike ipv6 address
         # where we require the full 128 even if some of them are just ':::'
         'ipv6_subnet': (r'(?:[a-fA-F0-9]{0,4}:){0,7}[a-fA-F0-9]{0,4}/[\d]{1,2}',
@@ -346,16 +348,23 @@ class Labels(JSONField):
         'asn': (r'[\d]+', "12345"),
         'vlan': (r'[\d]{1,4}', "1234"),
         'vlan_range': (r'[\d]{1,4}-[\d]{1,4}', "100-200"),
-        'inner_vlan': (r'[\d]{1,4}', "1234")
+        'inner_vlan': (r'[\d]{1,4}', "1234"),
+        'bgp_key': (r'[\w\-+_/\.:]{6,150}', "Amazon or other bgp key"),
+        # Azure UUID "3e2480b2-b4d5-3456-976a-7b0de65a1b62"
+        # GCP <random>/<vlan-attachment-region>/<edge-availability-domain> "7e51371e-1234-40b5-b844-2e3efefaee59/us-central1/2"
+        # AWS 12 digits
+        # Oracle ocid1.<RESOURCE TYPE>.<REALM>.[REGION][.FUTURE USE].<UNIQUE ID> "ocid1.instance.oc1.phx.abuw4ljrlsfiqw6vzzxb67hyypt4pkodawglp3wqxjqofakrwvou52gb6s5a"
+        'account_id': (r'[\w\-/\.]{3,100}', "Azure/GCP/AWS/Oracle account"),
+        'region': (r'[\w\-\.]{3,100}', "Azure/GCP/AWS/Oracle region identifier")
     }
     LAMBDA_VALIDATORS = {
-        'vlan': ((lambda v: True if 0 < int(v) <= 4096 else False), "1-4096"),
-        'inner_vlan': ((lambda v: True if 0 < int(v) <= 4096 else False), "1-4096"),
-        'vlan_range': ((lambda v: True if 0 < int(v.split('-')[0]) <= 4096 and
-                                         0 < int(v.split('-')[1]) <= 4096 and
-                                         int(v.split('-')[0]) < int(v.split('-')[1]) else False),
-                       "1-4096"),
-        'asn': ((lambda a: True if 0 < int(a) < 65536 else False), "1-65535")
+        'vlan': ((lambda v: True if 0 <= int(v) <= 4096 else False), "0-4096"),
+        'inner_vlan': ((lambda v: True if 0 <= int(v) <= 4096 else False), "0-4096"),
+        'vlan_range': ((lambda v: True if 0 <= int(v.split('-')[0]) <= 4096 and
+                                          0 <= int(v.split('-')[1]) <= 4096 and
+                                          int(v.split('-')[0]) <= int(v.split('-')[1]) else False),
+                       "0-4096"),
+        'asn': ((lambda a: True if 0 < int(a) < 2**32 else False), "1-4294967295")
     }
 
     def __init__(self, **kwargs):
@@ -376,6 +385,9 @@ class Labels(JSONField):
         self.local_name = None
         self.local_type = None
         self.device_name = None
+        self.bgp_key = None
+        self.account_id = None
+        self.region = None
         self._set_fields(**kwargs)
 
     def _set_fields(self, **kwargs):
@@ -556,6 +568,39 @@ class Location(JSONField):
         return float(response_json[0]['lat']), float(response_json[0]['lon'])
 
 
+class Flags(JSONField):
+    """
+    JSON-ified representation of various flags that can be attached to slivers
+    """
+    def __init__(self, **kwargs):
+        self.auto_config = False # primarily for interfaces
+        self.auto_mount = False # primarily for storage components
+        self.ipv4_management = False # request ipv4 management IP
+        self._set_fields(**kwargs)
+
+    def _set_fields(self, **kwargs):
+        for k, v in kwargs.items():
+            assert v is not None
+            assert isinstance(v, bool)
+            try:
+                # will throw exception if field is not defined
+                self.__getattribute__(k)
+                self.__setattr__(k, v)
+            except AttributeError:
+                raise FlagException(f"Unable to set field {k} of flags, no such field available")
+        return self
+
+    def to_json(self) -> str:
+        """
+        Dumps to JSON the __dict__ of the instance. Be careful as the fields in this
+        class should only be those that can be present in JSON output.
+        Specialized for boolean values from the generic implementation in JSONFields
+        :return:
+        """
+        d = self.__dict__.copy()
+        return json.dumps(d, skipkeys=True, sort_keys=True)
+
+
 class CapacityException(Exception):
     """
     Exception with a capacity
@@ -609,3 +654,12 @@ class LocationException(Exception):
     def __init__(self, msg: str):
         assert msg is not None
         super().__init__(f"Location exception: {msg}")
+
+
+class FlagException(Exception):
+    """
+    Exception with flags
+    """
+    def __init__(self, msg: str):
+        assert msg is not None
+        super().__init__(f"Flag exception: {msg}")
\ No newline at end of file
