@@ -29,8 +29,8 @@ from abc import ABC, abstractmethod
 import enum
 
 from ..graph.abc_property_graph import ABCPropertyGraph
-from ..slivers.capacities_labels import Capacities, Labels
-from ..slivers.measurement_data import MeasurementData
+from ..slivers.capacities_labels import Capacities, Labels, ReservationInfo
+from ..slivers.json_data import MeasurementData, UserData, LayoutData
 
 
 class ElementType(enum.Enum):
@@ -127,6 +127,12 @@ class ModelElement(ABC):
         :return:
         """
 
+    @abstractmethod
+    def get_sliver(self):
+        """
+        Return sliver of appropriate type for this element
+        """
+
     @property
     def name(self):
         return self._name
@@ -186,6 +192,15 @@ class ModelElement(ABC):
         if self.__dict__.get('topo', None) is not None:
             self.set_property('tags', value)
 
+    @property
+    def flags(self):
+        return self.get_property('flags') if self.__dict__.get('topo', None) is not None else None
+
+    @flags.setter
+    def flags(self, value):
+        if self.__dict__.get('topo', None) is not None:
+            self.set_property('flags', value)
+
     @property
     def mf_data(self):
         d = self.get_property('mf_data') if self.__dict__.get('topo', None) is not None else None
@@ -199,6 +214,50 @@ class ModelElement(ABC):
             else:
                 self.set_property('mf_data', MeasurementData(value))
 
+    @property
+    def user_data(self):
+        d = self.get_property('user_data') if self.__dict__.get('topo', None) is not None else None
+        return d.data if d is not None else None
+
+    @user_data.setter
+    def user_data(self, value):
+        if self.__dict__.get('topo', None) is not None:
+            if value is None or isinstance(value, UserData):
+                self.set_property('user_data', value)
+            else:
+                self.set_property('user_data', UserData(value))
+
+    @property
+    def layout_data(self):
+        d = self.get_property('layout_data') if self.__dict__.get('topo', None) is not None else None
+        return d.data if d is not None else None
+
+    @layout_data.setter
+    def layout_data(self, value):
+        if self.__dict__.get('topo', None) is not None:
+            if value is None or isinstance(value, LayoutData):
+                self.set_property('layout_data', value)
+            else:
+                self.set_property('layout_data', LayoutData(value))
+
+    @property
+    def boot_script(self):
+        return self.get_property('boot_script') if self.__dict__.get('topo', None) is not None else None
+
+    @boot_script.setter
+    def boot_script(self, value: str):
+        if self.__dict__.get('topo', None) is not None:
+            self.set_property('boot_script', value)
+
+    @property
+    def reservation_info(self):
+        return self.get_property('reservation_info') if self.__dict__.get('topo', None) is not None else None
+
+    @reservation_info.setter
+    def reservation_info(self, value: ReservationInfo):
+        if self.__dict__.get('topo', None) is not None:
+            self.set_property('reservation_info', value)
+
     def update_labels(self, **kwargs):
         """
         Take current labels field (even empty) and add/overwrite additional values into it
